/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (c) 2005, 2006 Tilghman Lesher
 *
 * Tilghman Lesher <func_odbc__200508@the-tilghman.com>
 * Steven Sokol <ssokol@sokol-associates.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 *
 * \brief ODBC multi-row lookups and row handling
 *
 * \author Tilghman Lesher <func_odbc__200508@the-tilghman.com>
 * \       Steven Sokol <ssokol@sokol-associates.com>
 */

/*** MODULEINFO
	<depend>unixodbc</depend>
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 40722 $")

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "asterisk/module.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/options.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/res_odbc.h"
#include "asterisk/app.h"

static char *config = "func_odbc_multi.conf";

enum {
	OPT_ESCAPECOMMAS =	(1 << 0),
} odbc_option_flags;

struct acf_odbc_query {
	AST_LIST_ENTRY(acf_odbc_query) list;
	char dsn[30];
	char sql_read[2048];
	unsigned int flags;
	struct ast_custom_function *acf;
};

AST_LIST_HEAD_STATIC(queries, acf_odbc_query);

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	int res;
	char *sql = data;
	SQLHSTMT stmt;

	res = SQLAllocHandle (SQL_HANDLE_STMT, obj->con, &stmt);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}

	res = SQLPrepare(stmt, (unsigned char *)sql, SQL_NTS);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Prepare failed![%s]\n", sql);
		SQLFreeHandle (SQL_HANDLE_STMT, stmt);
		return NULL;
	}

	return stmt;
}

static int acf_record_count(struct ast_channel *chan, char *cmd, char *s, char *buf, size_t len)
{
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	AST_STANDARD_APP_ARGS(args, s);
	if (args.argc < 1) {
		ast_log(LOG_ERROR, "First argument must be a CSV formatted string (i.e. recordset)\n");
		return -1;
	}
	int records = 0;
	char delim[] = "\x04";
	char *result = NULL;
	char *data = strdupa(args.field[0]);
	result = strtok(data, delim);
	while (result != NULL) {
		records++;
		result = strtok(NULL, delim);
	}

	if (records == 0)
		sprintf(buf, "0");
	else
		sprintf(buf, "%d", (records-1));
	return 0;
}

static int acf_record_fetch(struct ast_channel *chan, char *cmd, char *s, char *buf, size_t len)
{
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	AST_STANDARD_APP_ARGS(args, s);
	if (args.argc < 2) {
		ast_log(LOG_ERROR, "Function RECORD_FETCH requires two arguments.\n");
		return -1;
	}

	int i = 0;
	int record = 0;
	char delim[] = "\x04";
	char *result = NULL;
	char *fieldnames = NULL;

	char *data = strdupa(args.field[0]);
	char *recstring = strdupa(args.field[1]);

	int recnumber = strtol(recstring, NULL,10);
	
	result = strtok(data, delim);

	/* If the user wants this in RECORD format, save the field names */
	if (args.argc == 3) {
		if (!strcmp("ASSOC", args.field[2])) {
			fieldnames = strdupa(result);
		}
	}

	/* Now find the matching record and return it */
	while (result != NULL) {
		result = strtok(NULL, delim);
		record++;
		if (record == recnumber) {
			if (!fieldnames) {
				for (i = 0; i < strlen(result); i++) {
					if (result[i] == '\x03')
						result[i] = ',';
				}
				result[i - 1] = '\0';
				sprintf(buf, "%s", result);
			} else
				sprintf(buf, "%s\x04%s", fieldnames, result);
			return 0;
		}
	}
	ast_log(LOG_WARNING, "No record index matching your request was found\n");
	return -1;
}

static int acf_record(struct ast_channel *chan, char *cmd, char *s, char *buf, size_t len)
{
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	AST_STANDARD_APP_ARGS(args, s);
	if (args.argc < 2) {
		ast_log(LOG_ERROR, "Function RECORD requires two arguments.\n");
		return -1;
	}
	char delim[] = "\x04";
	char *data = strdupa(args.field[0]);
	char *field = strdupa(args.field[1]);
	/* First, get the rows (field names, field values) */
	char *fieldnames = strtok(data, delim);
	char *fieldvals = strtok(NULL, delim);

	if ((!fieldnames) || (!fieldvals)) {
		ast_log(LOG_ERROR, "Not a valid record object.\n");
		return -1;
	}
		
	/* Now, find the index of the matching field name */
	int index = -1;
	int  count = 0;
	char sep[] = "\x03";
	char *result = strtok(fieldnames, sep);
	while (result != NULL) {
		if (!strcmp(result, field)) {
			index = count;
			break;
		}
		count++;
		result = strtok(NULL, sep);
	}
	if (index == -1) {
		ast_log(LOG_ERROR, "The field name '%s' was not found.\n", field);
		return -1;
	}

	/* Now, find the value of the matching index in the data row */
	count = 0;
	char *fdata = strtok(fieldvals, sep);
	while (fdata != NULL) {
		if (count == index) {
			sprintf(buf, "%s", fdata);
			return 0;
		}
		count++;
		fdata = strtok(NULL, sep);
	}
	ast_log(LOG_ERROR, "The field name '%s' was present in the record.\n", field);
	return -1;
}

static int acf_field_count(struct ast_channel *chan, char *cmd, char *s, char *buf, size_t len)
{
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	AST_STANDARD_APP_ARGS(args, s);
	return 0;
}


/* RECORD_COUNT custom function */
static struct ast_custom_function record_count_function = {
	.name = "RECORD_COUNT",
	.synopsis = "Returns the number of records (rows) in a CSV recordset.",
	.syntax = "RECORD_COUNT(<csv string>)",
	.desc =
"Used with ODBC Multi-Row queries.  Returns the number\n"
"of records (rows) in a recordset.  For example:\n"
"count='${RECORD_COUNT(${MyRecordset})}'\n",
	.read = acf_record_count,
	.write = NULL,
};

/* RECORD_FETCH custom function */
static struct ast_custom_function record_fetch_function = {
	.name = "RECORD_FETCH",
	.synopsis = "Returns a specific record by row number (starting with 1).",
	.syntax = "RECORD_FETCH(<recordse string>,<row number>[,<format>])",
	.desc =
"Used with ODBC Multi-Row queries.  Returns a specific row\n"
"from the given recordset, optionally formatting it for pseudo-\n"
"associative access using the RECORD function. Example:\n"
"MyRecord='${RECORD_FETCH(${MyRecordset},1,ASSOC)}'\n"
"Format is comma separated if the ASSOC format paramter is omitted.\n",
	.read = acf_record_fetch,
	.write = NULL,
};

/* RECORD custom function */
static struct ast_custom_function record_function = {
	.name = "RECORD",
	.synopsis = "Allows for psudo-associative acccess to record data.",
	.syntax = "RECORD(<record string><field name>)",
	.desc =
"Used with ODBC Multi-Row queries.  Returns the value\n"
"of a field identified by field name.  For example:\n"
"MyVar='${RECORD(${MyRecord},field_name)}'\n"
"Note that RECORD only works with data formatted by the\n"
"RECORD_FETCH function with the ASSOC flag set.\n",
	.read = acf_record,
	.write = NULL,
};

/* FIELD_COUNT custom function */
static struct ast_custom_function field_count_function = {
	.name = "FIELD_COUNT",
	.synopsis = "Returns the number of fields (columns) in a CSV record.",
	.syntax = "FIELD_COUNT(<csv record>)",
	.desc =
"Used with ODBC Multi-Row queries.  Returns the number\n"
"of fields (columns) in a record.  For example:\n"
"count='${FIELD_COUNT(${MyRecord})}'\n",
	.read = acf_field_count,
	.write = NULL,
};

/*
 * Master control routine
 */

static int acf_odbc_read(struct ast_channel *chan, char *cmd, char *s, char *buf, size_t len)
{
	struct odbc_obj *obj;
	struct acf_odbc_query *query;
	char sql[2048] = "", varname[15];
	int res, x, r, buflen = 0, escapecommas;
	char eof[] = "\x03";
	char eol[] = "\x04";

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(field)[100];
	);
	SQLHSTMT stmt;
	SQLSMALLINT colcount=0;
	SQLINTEGER rowcount=0;
	SQLINTEGER indicator;

	SQLCHAR		colname[128];
	SQLSMALLINT	coltype;
	SQLSMALLINT	colnamelen;
	SQLUINTEGER	precision;
	SQLSMALLINT	scale;

	AST_LIST_LOCK(&queries);
	AST_LIST_TRAVERSE(&queries, query, list) {
		if (!strcmp(query->acf->name, cmd)) {
			break;
		}
	}

	if (!query) {
		ast_log(LOG_ERROR, "No such function '%s'\n", cmd);
		AST_LIST_UNLOCK(&queries);
		return -1;
	}

	obj = odbc_request_obj(query->dsn, 0);

	if (!obj) {
		ast_log(LOG_ERROR, "No such DSN registered (or out of connections): %s (check res_odbc.conf)\n", query->dsn);
		AST_LIST_UNLOCK(&queries);
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, s);
	for (x = 0; x < args.argc; x++) {
		snprintf(varname, sizeof(varname), "ARG%d", x + 1);
		pbx_builtin_pushvar_helper(chan, varname, args.field[x]);
	}

	pbx_substitute_variables_helper(chan, query->sql_read, sql, sizeof(sql) - 1);

	/* Restore prior values */
	for (x = 0; x < args.argc; x++) {
		snprintf(varname, sizeof(varname), "ARG%d", x + 1);
		pbx_builtin_setvar_helper(chan, varname, NULL);
	}

	/* Save this flag, so we can release the lock */
	escapecommas = ast_test_flag(query, OPT_ESCAPECOMMAS);

	AST_LIST_UNLOCK(&queries);

	stmt = odbc_prepare_and_execute(obj, generic_prepare, sql);

	if (!stmt) {
		odbc_release_obj(obj);
		return -1;
	}

	res = SQLRowCount(stmt, &rowcount);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Row Count error!\n[%s]\n\n", sql);
		SQLFreeHandle (SQL_HANDLE_STMT, stmt);
		odbc_release_obj(obj);
		return -1;
	}
	
	res = SQLNumResultCols(stmt, &colcount);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Column Count error!\n[%s]\n\n", sql);
		SQLFreeHandle (SQL_HANDLE_STMT, stmt);
		odbc_release_obj(obj);
		return -1;
	}

	*buf = '\0';
	char fieldnames[1024];
	char fieldvals[1024];

	/* Create row 0 with the column names */
	for (x = 0; x < colcount; x++) {
		SQLDescribeCol(stmt, x + 1, colname, sizeof(colname), &colnamelen, &coltype, &precision, &scale, NULL);
		
		if ((colnamelen + strlen(fieldnames)) > (len - 2)) {
			ast_log(LOG_WARNING, "Too much data for Function buffer.  Return truncated\n");
			return -1;
		} else {
			strcat((char *)fieldnames, colname);
			strcat((char *)fieldnames, eof);
		}
	}

	/* if our buffer is full, return an error */
	if (buflen >= (len - 2)) {
		ast_log(LOG_WARNING, "Too much data for Function buffer.  No rows can be returned\n");
		return -1;
	}

	strcat((char *)fieldnames, eol);
	strcat((char *)buf, (char *)fieldnames);


	for (r = 0; r < rowcount; r++) {
	
		res = SQLFetch(stmt);
		if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
			int res1 = -1;
			if (res == SQL_NO_DATA) {
				if (option_verbose > 3) {
					ast_verbose(VERBOSE_PREFIX_4 "Found no rows [%s]\n", sql);
				}
				res1 = 0;
			} else if (option_verbose > 3) {
				ast_log(LOG_WARNING, "Error %d in FETCH [%s]\n", res, sql);
			}
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			odbc_release_obj(obj);
			return res1;
		}
		
		fieldvals[0] = '\0';

		for (x = 0; x < colcount; x++) {
			char coldata[256];
	
			buflen = strlen(buf);
			res = SQLGetData(stmt, x + 1, SQL_CHAR, coldata, sizeof(coldata), &indicator);
			if (indicator == SQL_NULL_DATA) {
				coldata[0] = '\0';
				res = SQL_SUCCESS;
			}

			if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
				ast_log(LOG_WARNING, "SQL Get Data error!\n[%s]\n\n", sql);
				SQLFreeHandle(SQL_HANDLE_STMT, stmt);
				odbc_release_obj(obj);
				return -1;
			}

			strcat((char *)fieldvals, coldata);
			strcat((char *)fieldvals, eof);
		}

		if (r == rowcount)
			strcat((char *)fieldvals, "\0");
		else	
			strcat((char *)fieldvals, eol);

		if ((strlen(buf) + strlen(fieldvals)) > 4094) {
			ast_log(LOG_WARNING, "Appending record would exceed Function buffer.  Data truncated at record %d of %d.\n", (r - 1), (int)rowcount);
			break;
		} else {
			strcat((char *)buf, (char *)fieldvals);
		}
	}

	/* Lame C Code: What's the best way to get an integer into a string? */
	char qrows[10];
	char rrows[10];
	sprintf(qrows, "%d", (int)rowcount);
	sprintf(rrows, "%d", (r - 1));

	pbx_builtin_setvar_helper(chan, "QUERYCOUNT", qrows);
	pbx_builtin_setvar_helper(chan, "RETURNCOUNT", rrows);

	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	odbc_release_obj(obj);
	return 0;
}

static int init_acf_query(struct ast_config *cfg, char *catg, struct acf_odbc_query **query)
{
	char *tmp;

	if (!cfg || !catg) {
		return -1;
	}

	*query = ast_calloc(1, sizeof(struct acf_odbc_query));
	if (! (*query))
		return -1;

	if ((tmp = ast_variable_retrieve(cfg, catg, "dsn"))) {
		ast_copy_string((*query)->dsn, tmp, sizeof((*query)->dsn));
	} else {
		return -1;
	}

	if ((tmp = ast_variable_retrieve(cfg, catg, "read"))) {
		ast_copy_string((*query)->sql_read, tmp, sizeof((*query)->sql_read));
	}

	/* Allow escaping of embedded commas in fields to be turned off */
	ast_set_flag((*query), OPT_ESCAPECOMMAS);
	if ((tmp = ast_variable_retrieve(cfg, catg, "escapecommas"))) {
		if (ast_false(tmp))
			ast_clear_flag((*query), OPT_ESCAPECOMMAS);
	}

	(*query)->acf = ast_calloc(1, sizeof(struct ast_custom_function));
	if (! (*query)->acf) {
		free(*query);
		return -1;
	}

	if ((tmp = ast_variable_retrieve(cfg, catg, "prefix")) && !ast_strlen_zero(tmp)) {
		asprintf((char **)&((*query)->acf->name), "%s_%s", tmp, catg);
	} else {
		asprintf((char **)&((*query)->acf->name), "ODBC_MULTI_%s", catg);
	}

	if (!((*query)->acf->name)) {
		free((*query)->acf);
		free(*query);
		return -1;
	}

	asprintf((char **)&((*query)->acf->syntax), "%s(<arg1>[...[,<argN>]])", (*query)->acf->name);

	if (!((*query)->acf->syntax)) {
		free((char *)(*query)->acf->name);
		free((*query)->acf);
		free(*query);
		return -1;
	}

	(*query)->acf->synopsis = "Runs the referenced query with the specified arguments";
	if (!ast_strlen_zero((*query)->sql_read)) {
		asprintf((char **)&((*query)->acf->desc),
					"Runs the following query, as defined in func_odbc.conf, performing\n"
				   	"substitution of the arguments into the query as specified by ${ARG1},\n"
					"${ARG2}, ... ${ARGn}.  When setting the function, the values are provided\n"
					"either in whole as ${VALUE} or parsed as ${VAL1}, ${VAL2}, ... ${VALn}.\n"
					"\nRead:\n%s\n",
					(*query)->sql_read);
	}

	/* Could be out of memory, or could be we have neither sql_read nor sql_write */
	if (! ((*query)->acf->desc)) {
		free((char *)(*query)->acf->syntax);
		free((char *)(*query)->acf->name);
		free((*query)->acf);
		free(*query);
		return -1;
	}

	if (ast_strlen_zero((*query)->sql_read)) {
		(*query)->acf->read = NULL;
	} else {
		(*query)->acf->read = acf_odbc_read;
	}

	return 0;
}

static int free_acf_query(struct acf_odbc_query *query)
{
	if (query) {
		if (query->acf) {
			if (query->acf->name)
				free((char *)query->acf->name);
			if (query->acf->syntax)
				free((char *)query->acf->syntax);
			if (query->acf->desc)
				free((char *)query->acf->desc);
			free(query->acf);
		}
		free(query);
	}
	return 0;
}

static int odbc_load_module(void)
{
	int res = 0;
	struct ast_config *cfg;
	char *catg;

	AST_LIST_LOCK(&queries);

	cfg = ast_config_load(config);
	if (!cfg) {
		ast_log(LOG_NOTICE, "Unable to load config for func_odbc: %s\n", config);
		AST_LIST_UNLOCK(&queries);
		return 0;
	}

	for (catg = ast_category_browse(cfg, NULL);
	     catg;
	     catg = ast_category_browse(cfg, catg)) {
		struct acf_odbc_query *query = NULL;

		if (init_acf_query(cfg, catg, &query)) {
			ast_log(LOG_ERROR, "Out of memory\n");
			free_acf_query(query);
		} else {
			AST_LIST_INSERT_HEAD(&queries, query, list);
			ast_custom_function_register(query->acf);
		}
	}
	
	ast_config_destroy(cfg);
	ast_custom_function_register(&record_count_function);
	ast_custom_function_register(&record_fetch_function);
	ast_custom_function_register(&record_function);
	ast_custom_function_register(&field_count_function);

	AST_LIST_UNLOCK(&queries);
	return res;
}

static int odbc_unload_module(void)
{
	struct acf_odbc_query *query;

	AST_LIST_LOCK(&queries);
	while (!AST_LIST_EMPTY(&queries)) {
		query = AST_LIST_REMOVE_HEAD(&queries, list);
		ast_custom_function_unregister(query->acf);
		free_acf_query(query);
	}

	/* Allow any threads waiting for this lock to pass (avoids a race) */
	AST_LIST_UNLOCK(&queries);
	AST_LIST_LOCK(&queries);

	AST_LIST_UNLOCK(&queries);
	return 0;
}

static int reload(void)
{
	int res = 0;
	struct ast_config *cfg;
	struct acf_odbc_query *oldquery;
	char *catg;

	AST_LIST_LOCK(&queries);

	while (!AST_LIST_EMPTY(&queries)) {
		oldquery = AST_LIST_REMOVE_HEAD(&queries, list);
		ast_custom_function_unregister(oldquery->acf);
		free_acf_query(oldquery);
	}

	cfg = ast_config_load(config);
	if (!cfg) {
		ast_log(LOG_WARNING, "Unable to load config for func_odbc: %s\n", config);
		goto reload_out;
	}

	for (catg = ast_category_browse(cfg, NULL);
	     catg;
	     catg = ast_category_browse(cfg, catg)) {
		struct acf_odbc_query *query = NULL;

		if (init_acf_query(cfg, catg, &query)) {
			ast_log(LOG_ERROR, "Cannot initialize query %s\n", catg);
		} else {
			AST_LIST_INSERT_HEAD(&queries, query, list);
			ast_custom_function_register(query->acf);
		}
	}

	ast_config_destroy(cfg);
reload_out:
	AST_LIST_UNLOCK(&queries);
	return res;
}

static int unload_module(void)
{
	return odbc_unload_module();
}

static int load_module(void)
{
	return odbc_load_module();
}

/* XXX need to revise usecount - set if query_lock is set */

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "ODBC multi-row lookups",
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
	       );

